
/*--------------------------------------------------------------------.
| menus.c							      |
| 								      |
| Copyright (c) 1997 Jonathan Hudson				      |
| 								      |
| You may distribute under the terms of either the GNU General Public |
| License or the Artistic License, as specified in the README file.   |
`--------------------------------------------------------------------*/

static char rcsid[] = "$Id: menus.c,v 1.11 1999/05/07 17:43:13 jrh Exp $";

/*
 * $Log: menus.c,v $
 * Revision 1.11  1999/05/07 17:43:13  jrh
 * *** empty log message ***
 *
 * Revision 1.10  1998/03/15 20:28:01  jrh
 * updates for v1.09 and v1.10 (position + RC on command line)
 *
 * Revision 1.9  1997/11/15 14:33:02  jrh
 * reindented for release 1.05
 *
 * Revision 1.8  1997/11/14 22:37:48  jrh
 * optimised reversed menus
 *
 * Revision 1.7  1997/11/14 20:30:15  jrh
 * bidirectional menus
 *
 * Revision 1.6  1997/11/13 22:10:16  jrh
 * *** empty log message ***
 *
 * Revision 1.5  1997/10/12 20:41:40  jrh
 * indented source code
 *
 * Revision 1.4  1997/10/12 20:38:05  jrh
 * Added text for separators and changed the colours a bit
 *
 * Revision 1.3  1997/10/04 20:14:52  jrh
 * Yet more tidy ups
 *
 * Revision 1.1  1997/10/02 21:28:29  jrh
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <qdos.h>
#include <qptr.h>
#include <unistd.h>
#include "qascade.h"

typedef struct WM_menw WM_menw_t;

static long MEXEC (WM_wwork_t *, WM_menw_t *, char *, short, short);
static long MMENU (WM_wwork_t *, WM_menw_t *, char *, short, short);
static long MDRAW (WM_wwork_t *, WM_menw_t *);

static WM_action_t mdraw =
{JSR, wm_drwaw, MDRAW};
static WM_action_t mexec =
{JSR, wm_actme, MEXEC};
static WM_action_t mmenu =
{JSR, wm_actme, MMENU};

#define MAX_SECT    1
#define EXTRAS      1

static long MESC (WM_wwork_t * wwk, WM_litm_t * li);
static WM_action_t mesc =
{JSR, wm_actli, MESC};

static QD_TEXTI (ESCAPE, "Esc");

/*-----------------------------.
| * Reset a selected menu item |
`-----------------------------*/

static long lit_reset (WM_wwork_t * wwk, short lit)
{
    wwk->wstat->litem[lit] = WSI_MKAV;	/* set make available */
    return wm_ldraw (wwk, -1);	/* and re-draw */
}

/*------------------------------------------------.
| Called for us if the user presses ESC in a menu |
`------------------------------------------------*/

static long MESC (WM_wwork_t * wwk, WM_litm_t * li)
{
    lit_reset (wwk, li->item);
    return PT__CAN;
}

/*--------------------------------------------------------------------------.
|  Check if the menu fits in its window, and add scroll bars as necessary.  |
| Taken almost verbatim from the TT/c68 libqptr examples.		    |
`--------------------------------------------------------------------------*/

static void menu_chks (WM_menw_t * menu)
{
    WM_pwcb_t *pwcb;
    WM_splst_t *xl, *yl;

    short spacing;
    int i;

    pwcb = menu->ypwcb;		/* part window control block (scroll) */
    spacing = -menu->ys.c._spce;

    if (menu->nrow > menu->ysize / spacing)
    {
	if (pwcb->nsec == 0)
	    menu->xsize -= WW_XBAR;	/* adjust width for bar */
	pwcb->nsec = 1;		/* one section */
	pwcb->s[0].stt = 0;	/* from zero */
	pwcb->s[0].siz = ((menu->ysize - WW_YARROW * 2) / spacing);	/* size */
	menu->yoff = WW_YARROW;	/* offset for arrows */
    }
    else
    {
	if (pwcb->nsec)
	    menu->xsize += WW_XBAR;	/* adjust width for no bar */
	pwcb->nsec = 0;		/* no sections */
	menu->yoff = 0;		/* no arrows, no offset */
    }
    menu->xs.c._spce = -(menu->xsize) / menu->ncol;	/* width of each column */
    menu->xs.c._size = menu->xs.c._spce + menu->curw * 4;	/* ... item border */
}

/*---------------------------------------------------------.
|  Remove scroll bars from menu				   |
| Taken almost verbatim from the TT/c68 libqptr examples.  |
`---------------------------------------------------------*/

static void menu_rems (WM_menw_t * menu)
{
    WM_pwcb_t *pwcb;

    pwcb = menu->ypwcb;		/* part window control block (scroll) */

    if (pwcb->nsec)
    {
	menu->xsize += WW_XBAR;	/* adjust width for no bar */
	pwcb->nsec = 0;		/* no sections */
	menu->yoff = 0;
    }
}

/*-----------------------------------------------------------------------------.
| Load up a QPTR menu structure from our internal menu definition. Note that   |
| if the menu contains SEParators, then these built up here, as we didn't know |
| how wide they needed to be previously.				       |
`-----------------------------------------------------------------------------*/

static void LoadMenu (WM_menw_t * menu, MENUS * m)
{
    int i;
    WM_rowl_t *rowl;
    WM_mobj_t *mobj;
    char *mstt,skey = 0;
    int nitem;
    int nsize;
    ITEMS *im;

    if (menu->rowl)
    {
	xfree (menu->rowl);
    }

    menu->rowl = NULL;
    menu->ncol = 1;

    menu->nrow = nitem = m->height;

    nsize = (sizeof (WM_rowl_t) + sizeof (WM_mobj_t) +
	     sizeof (WM_mobj_t) + 1) * nitem;

    menu->rowl =		/* allocate all working space in a lump */
	rowl = (WM_rowl_t *) xalloc
	((int) ((sizeof (WM_rowl_t) + sizeof (WM_mobj_t) + 1) * nitem));
    mobj = (WM_mobj_t *) (((char *) rowl) + sizeof (WM_rowl_t) * nitem);
    menu->mstt =
	mstt = ((char *) mobj) + sizeof (WM_mobj_t) * nitem;

    im = m->items;
    for (i = 0; i < menu->nrow; i++)	/* initialise menu structures */
    {
	mobj->xjst = 1;
	mobj->yjst = 0;
	mobj->type = 0;
	mobj->skey = 0;
	mobj->item = i;
	if (im->l->act == TITLE)
	{
	    *mstt = WSI_AVBL;
	    mobj->pact = &mmenu;
	}
	else if (im->l->act == SEP)
	{
	    *mstt = WSI_UNAV;
	    if (im->l->len == 0)
	    {
		short j;
		MINFO *lp = im->l;

		im->l = xalloc (sizeof (MINFO) + m->width + EXTRAS);
		im->l->act = SEP;
		im->l->len = m->width + EXTRAS;
		if (*lp->chrs)
		{
		    short ns = strlen (lp->chrs);
		    im->l->chrs[0] = '-';
		    strcpy (im->l->chrs + 1, lp->chrs);
		    j = ns + 1;
		}
		else
		{
		    j = 0;
		}
		for (; j < m->width + EXTRAS; j++)
		{
		    im->l->chrs[j] = '-';
		}
		xfree (lp);
	    }
	}
	else
	{
	    *mstt = WSI_AVBL;
	    mobj->pact = &mexec;
	}

        if(koff && (*mstt == WSI_AVBL))
        {
            char c;
            static const char arry[] =
            {"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ#$*@"};
            
            c = (skey < sizeof(arry)-1) ? *(arry+skey) : 0;
            
            if(c)
            {
                mobj->skey = im->l->chrs[0] = c;
                mobj->type = -1;
                skey++;
            }
        }
        mstt++;
	mobj->pobj = &im->l->len;
	rowl->rows = mobj++;	/* one object per row */
	(rowl++)->rowe = mobj;
	im = im->next;		/* next text */
    }
}


int menu (WM_wwork_t * wwk, MENUS * m, void **mp)
{

    short pr_w = 0;
    WM_wstat_t *ws;
    WM_info_t *info;
    WM_infw_t *infw;
    WM_menw_t *mw;

    /*--------------------------------------------------------------.
    | We only really need 2 of these, however, passing data back up |
    | from menu action routines is quite difficult, so this is a    |
    | convenient hack to hold a pointer.			    |
    | 								    |
    | I don't recommend you really write programs like this	    |
    `--------------------------------------------------------------*/

    WM_appl_t appl[3];
    WM_wwork_t *ww;
    WM_litm_t *ppl;
    int rval;
    WM_PWCB (2) pwcy;
    QD_text_t *ptitle;
    short xsize, ysize;
    short xo, yo;
    static short dirn;
    WM_wsiz_t wl;

    /* Check how big the menu can be  */
    iop_flim (wwk->chid, -1, &wl);

    /*----------------------------------------------------------------.
    | dynamically allocate space for all the data structures required |
    | for PTR menus. This is painfully necessary as we're called      |
    | recursively.						      |
    `----------------------------------------------------------------*/

    ppl = xalloc (2 * sizeof (WM_litm_t));
    info = xalloc (3 * sizeof (WM_info_t));
    infw = xalloc (2 * sizeof (WM_infw_t));
    ww = xalloc (sizeof (WM_wwork_t));
    mw = xalloc (sizeof (WM_menw_t));
    ws = xalloc (sizeof (WM_wstat_t));

    /*-------------------------------.
    | Now set up the menu structures |
    `-------------------------------*/

    ppl->xsize = 24;
    ppl->ysize = 10;
    ppl->yorg = 2;
    ppl->type = TYP_TEXT;
    ppl->skey = 3;
    ppl->pobj = &ESCAPE;
    ppl->pact = &mesc;
    (ppl + 1)->xsize = (ppl + 1)->ysize = -1;

    memset (&pwcy, '\0', sizeof (pwcy));

    (info + 0)->ysize = 10;
    (info + 0)->xorg = 2;
    (info + 0)->type = TYP_TEXT;
    (info + 0)->attr.t.ink = 1;

    (info + 1)->xsize = (info + 0)->ysize = -1;

    (infw + 0)->ysize = 14;
    (infw + 0)->papr = 84;

    (infw + 1)->ysize = 10;
    (infw + 1)->xorg = 4;
    (infw + 1)->yorg = 2;
    (infw + 1)->papr = 7;
    (infw + 1)->pobl = info;

    (infw + 2)->xsize = (infw + 2)->ysize = -1;

    mw->xorg = 2;
    mw->yorg = 14;
    mw->borw = 1;
    mw->borc = colarray[COLBORD];
    mw->papr = colarray[COLPAPR];
    mw->draw = &mdraw;
    mw->hit = wm__mhit;
    mw->ctrl = wm__pnsc;

    mw->ypsac = colarray[COLPSAC];
    mw->ypsbc = colarray[COLPSBC];
    mw->ypssc = colarray[COLPSCC];

    mw->curw = 1;
    mw->curc = colarray[COLCURC];
    mw->uback = colarray[COLUBACK];
    mw->uink = colarray[COLUINK];
    mw->aback = colarray[COLABACK];
    mw->aink = colarray[COLAINK];
    mw->sback = colarray[COLSBACK];
    mw->sink = colarray[COLSINK];
    mw->ncol = 1;
    mw->ys.c._size = -10;
    mw->ys.c._spce = -12;

    (appl + 0)->pappw = (WM_appw_t *) mw;
    (appl + 1)->pappw = NULL;

    ww->flag = NO_WCLEAR + 2;
    ww->borw = 1;
    ww->borc = colarray[COLBORD];
    ww->papr = 84;
    ww->curw = 1;
    ww->curc = 0;
    ww->uback = 7;
    ww->uink = 4;
    ww->aback = 7;
    ww->aink = 0;
    ww->sback = 4;
    ww->sink = 0;

    ww->ninfo = 2;
    ww->ninob = 1;
    ww->nappl = 1;

    ww->pappl = ww->splst = appl;
    ww->pinfo = infw;
    ww->plitm = ppl;
    ww->nlitm = 1;

    ws->wmode = FORM_QL4;
    ws->wwork = ww;		/* cross references */
    ww->wstat = ws;

    ysize = m->height * 12;

    pr_w = strlen (m->title);
    ptitle = xalloc (pr_w + 8);
    ptitle->len = pr_w;
    memcpy (ptitle->chrs, m->title, pr_w);

    pr_w = (pr_w + 1) * 6;

    info[0].pobj = ptitle;
    info[0].xsize = pr_w - 6;	/* String length */

    /*-------------------------------------.
    | Don't exceed the screen size now...  |
    `-------------------------------------*/

    xsize = 4 + (m->width + EXTRAS) * 6;

    if (ysize > wl.ysize - 24)
    {
	ysize = 12 * ((wl.ysize - 24) / 12);
	xsize += 8;
    }

    if (xsize < pr_w + 42)
    {
	xsize = pr_w + 42;
    }

    infw[0].xsize = xsize;
    infw[1].xsize = pr_w;	/* outer wight window */

    (ppl + 0)->xorg = xsize - (ppl + 0)->xsize - 4;

    /*-------------------------------------------------------.
    | load the PTR menu from the menu defined by the rc file |
    `-------------------------------------------------------*/

    LoadMenu (mw, m);

    ww->xsize = xsize;
    mw->xsize = xsize - 4;

    mw->ysize = ysize;
    ww->ysize = 15 + mw->ysize;
    mw->nysc = MAX_SECT;
    mw->ypwcb = (WM_pwcb_t *) & pwcy;
    ww->xorg = 4;
    ww->yorg = 15 + 8;

    ws->citem = -1;		/* no current item  */
    ws->ciact = NULL;		/* or current item routine */
    if (mp)			/* Top level invocation */
    {
	xo = wwk->wstat->ptpsx;
	/*--------------------------------------------------------------.
	|   Hide away the pointer here for subsequent action routines.  |
	`--------------------------------------------------------------*/
	(appl + 2)->pappw = (WM_appw_t *) mp;
	if (xo + 2 * xsize > wl.xsize)
	    dirn = -1;
	else
	    dirn = 1;
    }
    else
    {
	if (dirn == -1)
	{
	    xo = wwk->xorg - ww->xsize / 2;
	    if (xo < 0)
		dirn = 1;
	}
	else if (dirn == 1)
	{
	    xo = wwk->xorg + wwk->xsize / 2;
	    if (xo > wl.xsize)
		dirn = -1;
	}
	/*------------------------------------------------------.
	| Pass on the callers data pointer to subsequent menus, |
	| it can only be set by one action routine	        |
	`------------------------------------------------------*/
	(appl + 2)->pappw = (wwk->pappl + 2)->pappw;
    }
    yo = wwk->wstat->ptpsy + 10;
    /*----------------------------.
    | And draw the menu (at last) |
    `----------------------------*/

    wm_pulld (ww, xo, yo);
    wm_wdraw (ww);

    /*--------------------------------------------------------------------.
    | Now read the pointer, we will either get a CAN is the user pressed  |
    | ESC, or a DO if an invocaable item was selected. This may have been |
    | cascaded back from a sub menu, via the action routine.		  |
    `--------------------------------------------------------------------*/
    while (!wm_rptr (ww))
    {
	if (ws->evnt & PT_CAN)
	{
	    rval = 0;
	    break;
	}

	if (ws->evnt & PT_DO)
	{
	    wwk->wstat->ptpsy = ww->wstat->ptpsy;
	    wwk->wstat->ptpsx = ww->wstat->ptpsx;
	    rval = PT__DO;
	    break;
	}
#if 0
/*--------------------------------------------------------------------.
| move is not yet implemented (and never will be ?). If it were, then |
| this would handle it.						      |
`--------------------------------------------------------------------*/
	if (ws->evnt & PT_WMOVE)
	{
	    menu_rems (mw);
	    LoadMenu (mw, m);
	    wm_swapp (ww, 0, -1);
	    sd_clear (ww->chid, -1);
	    MDRAW (ww, mw);
	}
#endif
    }

    /*-----------------------------------------------------------------.
    | The user has finished with the menu, we must clean up after them |
    `-----------------------------------------------------------------*/

    if (mw->rowl)
    {
	xfree (mw->rowl);
    }

    wm_cluns (ww);		/* throw it away */
    xfree (ppl);
    xfree (ptitle);
    xfree (info);
    xfree (infw);
    xfree (ww);
    xfree (mw);
    xfree (ws);

    return rval;
}

/*----------------------------.
| The draw callback for menus |
`----------------------------*/

static long MDRAW (WM_wwork_t * wwk, WM_menw_t * mw)
{

    menu_chks (mw);
    wm_index (wwk, (WM_swdef_t *) mw);
    wm_mdraw (wwk, (WM_swdef_t *) mw, 0);
}

/*-------------------------------------------------------.
| Reset an item in a menu after the user has selected it |
`-------------------------------------------------------*/

static void resetobj (WM_wwork_t * wwk, WM_menw_t * mw, char *mst, short item)
{
    mst[item] = WSI_MKAV;
    wm_mdraw (wwk, (WM_swdef_t *) mw, -1);
    mst[item] = WSI_AVBL;
}

/*----------------------------------------------------------------------.
| This callback is called by the PE when the user clicks on a invocable |
| menu item.							        |
`----------------------------------------------------------------------*/

static long MEXEC (WM_wwork_t * wwk, WM_menw_t * mw, char *mst, short col,
		   short row)
{
    WM_mobj_t *mobj;
    void *q;
    char *s;

    mobj = mw->rowl[row].rows;	/* current menu object */
    q = mobj->pobj;
    s = ((char *) q - offsetof (MINFO, len));	/* Return the info object */
    /*-------------------------------------------------------------.
    | This is where we've stashed away the pointer to the original |
    | callers data.						   |
    `-------------------------------------------------------------*/
    *(void **) ((wwk->pappl + 2)->pappw) = s;
    resetobj (wwk, mw, mst, mobj->item);
    return PT__DO;
}

/*-------------------------------------------------------------------------.
| This callback is called by the PE when the user clicks on a menu	   |
| item that invokes another menu. As the callback in effectively called	   |
| from a previous call to menu(), then this is recursive (see recursion).  |
`-------------------------------------------------------------------------*/

static long MMENU (WM_wwork_t * wwk, WM_menw_t * mw, char *mst, short col,
		   short row)
{
    WM_mobj_t *mobj;
    void *q;
    MENUS *s;
    int res;

    mobj = mw->rowl[row].rows;	/* current menu object */
    q = mobj->pobj;
    s = *(MENUS **) (char *) ((char *) q - offsetof (MINFO, len));
    res = menu (wwk, s, NULL);
    resetobj (wwk, mw, mst, mobj->item);
    return res;
}
